package MyApp::Controller;
use strict;
use warnings;

use Router::Boom::Method;
use FormValidator::Lite;

use MyApp::Model::User;
use MyApp::Model::UserItems;
use MyApp::Model::Friends;
use MyApp::Model::Friend;
use JSON;

use Mouse;

has router => (
    is => "ro",
    isa => "Router::Boom::Method",
    default => sub {
        Router::Boom::Method->new,
    },
);

has view => (
    is => "ro",
    isa => "Object",
    default => sub {
        JSON->new->utf8->pretty;
    },
);

has resp_headers => (
    is => "ro",
    isa => "ArrayRef",
    default => sub {
        return [
            "Content-Type" => "application/json",
        ];
    },
);

no Mouse;

sub build {
    my $self = shift;
    my $router = $self->router;

    $router->add("POST", "/user", "register");
    $router->add("POST", "/user/:user_id/friend", "friend");
    $router->add("GET", "/user/:user_id/friends", "friends");
    $router->add("POST", "/my_user/item/add", "add_item");
    $router->add("POST", "/my_user/item/consume", "consume_item");
}

sub dispatch {
    my ($self, $req) = @_;
    my $router = $self->router;

    my ($dest, $captured, $is_method_not_allowed) =
        $router->match($req->method, $req->path_info);
    if ($is_method_not_allowed) {
        return $self->wrap("405", { error => "Method Not Allowed" });
    }
    my $status = "200";
    my $resp = eval {
        $self->$dest($req, $captured);
    };
    if (my $err = $@) {
        warn $err;
        ($status, $resp) = $self->error($err);
    }

    return $self->wrap($status, $resp);
}

sub error {
    my ($self, $err) = @_;

    return (500, { error => $err });
}

sub wrap {
    my ($self, $status, $resp) = @_;

    my $serialized = $self->view->encode($resp);
    my $len = length($serialized);

    return [
        $status,
        [
            @{$self->resp_headers},
            "Content-Length" => $len,
        ],
        [$serialized],
    ];
}

sub register {
    my ($self, $req) = @_;

    my $v = FormValidator::Lite->new($req);
    my $res = $v->check(
        name      => [qw/NOT_NULL/],
        avatar_id => [qw/NOT_NULL UINT/],
    );
    if ($res->has_error) {
        die { parameters => $res->errors };
    }

    my $name = $req->parameters->{name};
    my $avatar_id = $req->parameters->{avatar_id};

    my $user = MyApp::Model::User->new;
    $user->build(name => $name, avatar_id => $avatar_id);

    return $user->info_with_session;
}

sub friend {
    my ($self, $req, $captured) = @_;

    my $session = $req->parameters->{session};
    my $user = MyApp::Model::User->new_by_session($session);

    my $user_id = $captured->{user_id};
    my $friend_user = MyApp::Model::User->new_by_user_id($user_id);
    my $friend_rel = MyApp::Model::Friend->new;
    $friend_rel->build(user_a => $user, user_b => $friend_user);

    return $friend_rel->info;
}

sub friends {
    my ($self, $req, $captured) = @_;

    my $user_id = $captured->{user_id};

    my $user = MyApp::Model::User->new_by_user_id($user_id);
    my $friends = MyApp::Model::Friends->new_by_user($user);

    return $friends->info;
}

sub add_item {
    my ($self, $req) = @_;

    my $v = FormValidator::Lite->new($req);
    my $res = $v->check(
        session => [qw/NOT_NULL/],
        item_id => [qw/NOT_NULL UINT/],
        num     => [qw/NOT_NULL UINT/],
    );
    if ($res->has_error) {
        die { parameters => $res->errors };
    }

    my $session = $req->parameters->{session};
    my $item_id = $req->parameters->{item_id};
    my $num = $req->parameters->{num};

    my $user = MyApp::Model::User->new_by_session($session);
    my $user_items = MyApp::Model::UserItems->new_by_user($user);

    my $user_item = $user_items->add(item_id => $item_id, num => $num);

    return $user_item->info;
}

sub consume_item {
    my ($self, $req) = @_;

    my $v = FormValidator::Lite->new($req);
    my $res = $v->check(
        session => [qw/NOT_NULL/],
        item_id => [qw/NOT_NULL UINT/],
        num     => [qw/NOT_NULL UINT/],
    );
    if ($res->has_error) {
        die { parameters => $res->errors };
    }

    my $session = $req->parameters->{session};
    my $item_id = $req->parameters->{item_id};
    my $num     = $req->parameters->{num};

    my $user = MyApp::Model::User->new_by_session($session);
    my $user_items = MyApp::Model::UserItems->new_by_user($user);
    my $user_item = $user_items->user_item_by_item_id($item_id);

    my $result = $user_item->consume(num => $num);

    return $result;
}

1;
