package MyApp::Model::UserItem;
use strict;
use warnings;

use MyApp::Model::Item;
use Mouse;

extends "MyApp::Model::Base";

has user => (
    is => "rw",
    isa => "MyApp::Model::User",
);

has item_id => (
    is => "rw",
    isa => "Int",
);

has having_num => (
    is => "rw",
    isa => "Int",
);

has _item => (
    is => "rw",
    isa => "Maybe[MyApp::Model::Item]",
    default => undef,
);

no Mouse;

sub build {
    my ($self, %args) = @_;
    my $user = $args{user};
    my $item_id = $args{item_id};

    $self->user($user);
    $self->item_id($item_id);

    $self->_build;
    $self->_fetch;
}

sub _build {
    my $self = shift;

    $self->dbh->query(<<SQL, $self->user->id, $self->item_id, 0);
INSERT INTO `user_items` (user_id, item_id, having_num)
    VALUES (?, ?, ?)
    ON DUPLICATE KEY UPDATE having_num = having_num;
SQL
}

sub _fetch {
    my $self = shift;

    return if $self->_fetched;

    my $row = $self->dbh->select_row(<<SQL, $self->user->id, $self->item_id);
SELECT * FROM user_items WHERE user_id = ? AND item_id = ?;
SQL

    $self->__update_attr($row);
    $self->_fetched(1);
}

sub __update_attr {
    my ($self, $row) = @_;

    $self->$_($row->{$_}) for qw/item_id having_num/;
}

sub _refetch {
    my $self = shift;

    $self->_fetched(0);
    $self->_fetch;
}

sub info {
    my $self = shift;

    return {
        user_item => {
            user_id    => $self->user->id,
            item_id    => $self->item_id,
            having_num => $self->having_num,
        },
    };
}

sub add {
    my ($self, %args) = @_;
    my $num = $args{num};

    $self->dbh->query(<<SQL, $num, $self->user->id, $self->item_id);
UPDATE `user_items` SET having_num = having_num + ?
    WHERE user_id = ? AND item_id = ?
SQL
    $self->_refetch;

}

sub consume {
    my ($self, %args) = @_;
    my $num = $args{num};

    my $txn = $self->dbh->txn_scope;
    $self->user->lock;
    $self->lock;
    $self->_try_having_by_num($num);
    $self->_subtract(num => $num);
    my $result = $self->effect(num => $num);

    $txn->commit;

    return {
        %$result,
        %{$self->info},
    };
}

sub lock {
    my $self = shift;

    my $row = $self->dbh->select_row(<<SQL, $self->user->id, $self->item_id);
SELECT * FROM user_items WHERE user_id = ? AND item_id = ? FOR UPDATE;
SQL

    $self->__update_attr($row);
}

sub _try_having_by_num {
    my ($self, $num) = @_;

    if ($self->having_num < $num) {
        die "所持数より大きい数量を減らそうとしています";
    }
}

sub _subtract {
    my ($self, %args) = @_;
    my $num = $args{num};

    $self->dbh->query(<<SQL, $num, $self->user->id, $self->item_id);
UPDATE `user_items` SET having_num = having_num - ?
    WHERE user_id = ? AND item_id = ?;
SQL

    $self->_refetch;
}

sub effect {
    my ($self, %args) = @_;
    my $num = $args{num};

    my $effector = $self->_effector;
    my $result = $effector->effect(
        user => $self->user,
        num => $num,
    );

    return $result;
}

sub _effector {
    my $self = shift;

    my $item = $self->item;
    return $item->effector;
}

sub item {
    my $self = shift;

    return $self->_item if $self->_item;

    my $item = MyApp::Model::Item->new;
    $item->build(item_id => $self->item_id);
    $self->_item($item);

    return $item;
}

1;
